﻿


' ==================================================================================================
'   CLASS MASTER
' ==================================================================================================

Partial Friend Class Master

    Private ActionLock As Object = New Object
    Private Name As String
    Private mSlaveCount As Int32

    Friend Hid As Theremino_HID
    Friend ConfigId As Int32
    Friend ConfigValid As Boolean
    Friend Slaves As List(Of Slave)
    Friend MasterId As Int32

    Friend CommSpeed As Int32
    Friend FastComm As Boolean
    Friend CommMillisec As Single
    Friend CommFps As Single
    Friend ErrorRate As Single


    Friend ReadOnly Property SlaveCount() As Int32
        Get
            Return mSlaveCount
        End Get
    End Property

    Friend Sub SetName(ByVal _name As String)
        Name = _name
    End Sub

    Friend Function GetName() As String
        Return Name
    End Function

    Friend Sub New(ByVal _masterid As Int32)
        Name = "NoName"
        MasterId = _masterid
        mSlaveCount = 0
        CommSpeed = 7
        FastComm = False
        CommFps = 100
        Slaves = New List(Of Slave)
    End Sub

    'Protected Overrides Sub Finalize()
    '    If CanUseWindowsDlls() Then
    '        '
    '    Else
    '        '    If Hid.HidDevice_Unix.IsHidOpen() Then
    '        '        Hid.HidDevice_Unix.HidClose()
    '        '    End If
    '    End If
    'End Sub



    ' ==========================================================================================
    '   HOST TO MASTER COMMANDS
    ' ==========================================================================================

    Friend Enum Cmd As Byte
        ' --------------------------------------- commands that master expands for the slaves
        RECOG_START = 254
        'CMD_RECOG 						= 253 ( Master to Slaves only )
        FAST_DATA_EXCHANGE = 251
        '
        ' --------------------------------------- commands that master repeats to slaves
        SETUP_SLAVE_PINS = 249
        SET_MASTER_NAME = 248
        GET_MASTER_NAME = 247
        SEND_VALUES = 246
        GET_VALUES = 245
        SEND_BYTES = 244
        GET_BYTES = 243
        '
        ' --------------------------------------- commands to master only  ( < 200 )
        SPEED = 199
        NOCMD = 0
    End Enum




    ' ==========================================================================================
    '   PUBLIC COMMANDS
    ' ==========================================================================================
    Friend Sub Calibrate()
        If Not CalibrationNeeded Then Exit Sub
        DataExchangeThread_Stop()
        SyncLock ActionLock
            SleepMyThread(10)
            DataExchange()
            SleepMyThread(10)
            For i As Int32 = 0 To SlaveCount - 1
                Slaves(i).Calibrate()
            Next
        End SyncLock
        DataExchangeThread_Start()
    End Sub

    Friend Sub CalibrateZero()
        For i As Int32 = 0 To SlaveCount - 1
            Slaves(i).CalibrateZero()
        Next
    End Sub

    Friend Function CalibrationNeeded() As Boolean
        For i As Int32 = 0 To SlaveCount - 1
            If Slaves(i).CalibrationNeeded Then
                Return True
            End If
        Next
        Return False
    End Function

    Friend Sub SetSpeed()
        DataExchangeThread_Stop()
        SyncLock ActionLock
            SendSpeedToSlaves()
        End SyncLock
        DataExchangeThread_Start()
    End Sub

    Friend Sub SetSpeed(ByVal _Speed As Int32)
        DataExchangeThread_Stop()
        SyncLock ActionLock
            CommSpeed = _Speed
            SendSpeedToSlaves()
        End SyncLock
        DataExchangeThread_Start()
    End Sub


    Friend Sub Recognize()

        '' ------------------------------------------------- try to unlock slaves
        'Dim b(60) As Byte
        'SendBytesToSlave(255, b)
        'SleepMyThread(100)
        '' -------------------------------------------------

        DataExchangeThread_Stop()
        SyncLock ActionLock
            Slaves.Clear()
            ' --------------------------------------------- SPEED - TODO provare con molti slave e lasciarne solo due
            SendSpeedToSlaves()
            SendSpeedToSlaves()
            'SendSpeedToSlaves()
            'SendSpeedToSlaves()
            'SendSpeedToSlaves()
            'SendSpeedToSlaves()
            ' --------------------------------------------- DETECT
            Hid.USB_TxData(0) = Cmd.RECOG_START
            Hid.USB_TxData(1) = 0
            ' --------------------------------------------- WRITE-READ
            Hid.USB_Write_Read()
            Hid.USB_TxData(0) = Cmd.NOCMD
            ' --------------------------------------------- ERROR ?
            If Hid.USB_RxData(0) <> 0 Then
                Execution_Error("Recognize")
            Else
                Execution_OK()
            End If
            ' --------------------------------------------- INIT
            mSlaveCount = Hid.USB_RxData(1)
            For i As Int32 = 0 To mSlaveCount - 1
                Dim s As Slave.SlaveTypes = CType(Hid.USB_RxData(i + 2), Slave.SlaveTypes)
                Select Case s
                    Case Slave.SlaveTypes.CapSensor
                        Slaves.Add(New Slave_CapSensor(MasterId, i))
                    Case Slave.SlaveTypes.Servo
                        Slaves.Add(New Slave_Servo(MasterId, i))
                    Case Slave.SlaveTypes.Generic
                        Slaves.Add(New Slave_Generic(MasterId, i))
                    Case Slave.SlaveTypes.InOut
                        Slaves.Add(New Slave_InOut(MasterId, i))
                    Case Slave.SlaveTypes.MasterPins
                        Slaves.Add(New Slave_MasterPins(MasterId, i))
                    Case Slave.SlaveTypes.RmsLogger
                        Slaves.Add(New Slave_RmsLogger(MasterId, i))
                    Case Slave.SlaveTypes.Sniffer
                        Slaves.Add(New Slave_Sniffer(MasterId, i))
                    Case Else
                        mSlaveCount = 0
                        Slaves.Clear()
                        Exit For
                End Select
            Next
        End SyncLock
        DataExchangeThread_Start()
    End Sub

    Friend Sub SetupSlavePins(ByVal slaveID As Int32)
        DataExchangeThread_Stop()
        SyncLock ActionLock
            Dim s As Slave = Slaves(slaveID)
            Hid.USB_TxData(0) = Cmd.SETUP_SLAVE_PINS
            Hid.USB_TxData(1) = CByte(slaveID)
            Hid.USB_TxData(2) = CByte(s.Pins.Count)
            For i As Int32 = 0 To s.Pins.Count - 1
                Hid.USB_TxData(i + 3) = CByte(s.Pins(i).GetPinType)
            Next
            Hid.USB_Write_Read()
            Hid.USB_TxData(0) = Cmd.NOCMD
            ' --------------------------------------------- error ?
            If Hid.USB_RxData(0) <> 0 Then
                Execution_Error("SetupSlavePins")
            Else
                Execution_OK()
            End If
            ' ---------------------------------------------
            SetCommunicationIndexesToPins()
        End SyncLock
        DataExchangeThread_Start()
    End Sub

    Friend Sub WriteNameToHardware()
        DataExchangeThread_Stop()
        SyncLock ActionLock
            Name = Left(Name, 16)
            Hid.USB_TxData(0) = Cmd.SET_MASTER_NAME
            Dim i As Int32
            For i = 1 To Name.Length
                Hid.USB_TxData(i) = CByte(Asc(Mid(Name, i, 1)))
            Next
            Hid.USB_TxData(i) = CByte(0)
            Hid.USB_Write_Read()
            Hid.USB_TxData(0) = Cmd.NOCMD
            ' --------------------------------------------- error ?
            If Hid.USB_RxData(0) <> 0 Then
                Execution_Error("WriteNameToHardware")
            Else
                Execution_OK()
            End If
        End SyncLock
        DataExchangeThread_Start()
    End Sub

    Friend Sub ReadNameFromHardware()
        DataExchangeThread_Stop()
        SyncLock ActionLock
            Name = ""
            Hid.USB_TxData(0) = Cmd.GET_MASTER_NAME
            Hid.USB_Write_Read()
            Hid.USB_TxData(0) = Cmd.NOCMD
            ' --------------------------------------------- error ?
            If Hid.USB_RxData(0) = 0 Then
                Dim i As Int32 = 1
                Do
                    If Hid.USB_RxData(i) = 0 Then Exit Do
                    Name += Chr(Hid.USB_RxData(i))
                    i += 1
                Loop
                Execution_OK()
            Else
                Execution_Error("ReadNameToHardware")
            End If
        End SyncLock
        DataExchangeThread_Start()
    End Sub



    Friend Sub SetOutputPinsToParkingPosition()
        DataExchangeThread_Stop()
        SyncLock ActionLock
            For Each s As Slave In Slaves
                For Each p As Pin In s.Pins
                    If p.Direction = Pin.Directions.HostToMaster Then
                        p.Value = p.Value_Parking
                        p.ConvertValueToHardwareFormat()
                    End If
                Next
            Next
            ' ------------------------------------------------- USB <> Master <> Slaves
            DataExchange()
            SleepMyThread(32)
        End SyncLock
        DataExchangeThread_Start()
    End Sub

    Friend Sub StopAllSnifferLogs()
        For Each s As Slave In Slaves
            For Each p As Pin In s.Pins
                p.SnifferLogEnabled = False
            Next
        Next
    End Sub


    Friend Sub SetCommunicationIndexesToPins()
        ' ---------------------------------------------------- byte 0 is the command
        Dim UsbByteIndex_HostToHardware As Int32 = 1
        ' ---------------------------------------------------- byte 0 is the response
        Dim UsbByteIndex_HardwareToHost As Int32 = 1
        '
        For Each s As Slave In Slaves
            For Each p As Pin In s.Pins
                ' --------------------------------------------------- USB communication byte index
                Select Case p.Direction
                    Case Pin.Directions.HostToMaster
                        p.UsbByteIndex = UsbByteIndex_HostToHardware
                        UsbByteIndex_HostToHardware += p.UsbBytesCount
                    Case Pin.Directions.MasterToHost
                        p.UsbByteIndex = UsbByteIndex_HardwareToHost
                        UsbByteIndex_HardwareToHost += p.UsbBytesCount
                End Select
                'Debug.Print(p.PinId.ToString & "  " & Pin.PinTypeToString(p.GetPinType) & "  " & p.UsbByteIndex)
            Next
        Next
    End Sub



    ' ==========================================================================================
    '   INTERNAL COMMANDS
    ' ==========================================================================================
    Private Sub SendSpeedToSlaves()
        '
        Hid.USB_TxData(0) = Cmd.SPEED
        Hid.USB_TxData(1) = CByte(CommSpeed)
        '
        Hid.USB_Write_Read()
        Hid.USB_TxData(0) = Cmd.NOCMD
        ' --------------------------------------------- error ?
        If Hid.USB_RxData(0) <> 0 Then
            Execution_Error("SendSpeedToSlaves")
        Else
            Execution_OK()
        End If
    End Sub


    ' ==========================================================================================
    '   SMOOTHING FUNCTIONS
    ' ==========================================================================================
    Friend Sub SmoothValue(ByRef value As Single, _
                           ByVal new_value As Single, _
                           ByVal speed As Single, _
                           ByVal pow_speed As Boolean)
        '
        If Single.IsNaN(new_value) Then
            value = new_value
            Return
        End If
        ' --------------------------------------------------- exclude gradually the smoothing 
        Select Case speed
            Case 98 : speed = 150
            Case 99 : speed = 200
            Case 100 : speed = 300
        End Select
        '
        If pow_speed Then
            ' ------------------------------------------------ correction to have a similar response
            speed *= 10 '27
            '
            Dim delta As Single = new_value - value
            Dim delta_sgn As Single = Math.Sign(delta)
            Dim delta_abs As Single = Math.Abs(delta)

            Dim delta_pow As Single = CSng(delta_abs ^ 2) * speed / CommFps
            If value <> 0 Then delta_pow /= Math.Abs(value)

            If delta_pow >= delta_abs Then
                value = new_value
            Else
                value += delta_sgn * delta_pow
            End If
        Else
            Dim delta As Single = new_value - value
            Dim delta_sgn As Single = Math.Sign(delta)
            Dim delta_abs As Single = Math.Abs(delta)
            Dim delta_pow As Single = CSng(speed / CommFps * delta_abs)
            If delta_pow >= delta_abs Then
                value = new_value
            Else
                value += delta_sgn * delta_pow
            End If
        End If
    End Sub

    Private Sub SmoothValue_Pow(ByRef value As Single, ByVal new_value As Single, ByVal speed As Single)
        If Single.IsNaN(new_value) Then
            value = new_value
            Return
        End If

        Dim delta As Single = new_value - value
        Dim delta_sgn As Single = Math.Sign(delta)
        Dim delta_abs As Single = Math.Abs(delta)
        Dim delta_pow As Single
        delta_pow = CSng((speed / CommFps * delta_abs) ^ 2)

        If delta_pow >= delta_abs Then
            value = new_value
        Else
            value += delta_sgn * delta_pow
        End If
    End Sub


    ' ==========================================================================================
    '   LOOPS AND COMMUNICATIONS
    ' ==========================================================================================
    Private sw1 As Diagnostics.Stopwatch = New Diagnostics.Stopwatch

    Private Sub DataExchange()
        'CommMillisec = CSng(sw1.Elapsed.TotalMilliseconds - 0.1)
        'CommMillisec = CSng(sw1.Elapsed.TotalMilliseconds)
        CommMillisec = CSng(sw1.ElapsedMilliseconds)
        sw1.Reset()
        sw1.Start()
        If CommMillisec >= 1 Then
            SmoothValue_Pow(CommFps, 1000 / CommMillisec, 5)
        End If
        Try
            If FastComm Then
                FastDataExchange()
            Else
                SlowDataExchange()
            End If
        Catch
#If DEBUG Then
            Beep_RepetitionLimited(150)
            'MsgBox("Exception in ""DataExchange""")
            Debug.Print("Exception in ""DataExchange""")
#End If
        End Try
    End Sub


    Private Sub SlowDataExchange()
        Dim ix As Int32
        Dim b() As Byte
        Dim ExchangedFlag As Boolean = False
        ' -------------------------------------------------------- MMF to USB
        For Each s As Slave In Slaves
            Dim nbytes As Int32 = s.GetHostToMasterBytesCount
            If nbytes > 0 Then
                ix = 0
                ReDim b(nbytes - 1)
                For Each p As Pin In s.Pins
                    If p.Direction = Pin.Directions.HostToMaster Then
                        p.ReadValueFromMMF()
                        p.ConvertValueToHardwareFormat()
                        Select Case p.GetPinType
                            Case Pin.PinTypes.DIG_OUT
                                b(ix) = CByte(p.Value_RawInteger)
                                ix += 1
                            Case Pin.PinTypes.PWM_8, Pin.PinTypes.SERVO_8
                                b(ix) = CByte(p.Value_RawInteger / 256)
                                ix += 1
                            Case Pin.PinTypes.PWM_16, Pin.PinTypes.SERVO_16
                                Dim bytes() As Byte = ByteArrayFromUint16(CUShort(p.Value_RawInteger))
                                b(ix) = bytes(0)
                                ix += 1
                                b(ix) = bytes(1)
                                ix += 1
                        End Select
                    End If
                Next
                SendValuesToSlave(s.mSlaveId, b)
                ExchangedFlag = True
            End If
        Next
        ' -------------------------------------------------------- USB to MMF
        For Each s As Slave In Slaves
            Dim nbytes As Int32 = s.GetMasterToHostBytesCount
            If nbytes > 0 Then
                ExchangedFlag = True
                ix = 0
                ReDim b(nbytes - 1)
                If GetValuesFromSlave(s.mSlaveId, b) Then
                    For Each p As Pin In s.Pins
                        If p.Direction = Pin.Directions.MasterToHost Then
                            Select Case p.UsbBytesCount
                                Case 1
                                    p.Value_RawInteger = b(ix)
                                    ix += 1
                                Case 2
                                    p.Value_RawInteger = b(ix) * 256UI + b(ix + 1)
                                    ix += 2
                                Case 3
                                    p.Value_RawInteger = b(ix) * 65536UI + b(ix + 1) * 256UI + b(ix + 2)
                                    ix += 3
                                Case 4
                                    p.Value_RawInteger = b(ix) * 256UI * 65536UI + b(ix + 1) * 65536UI + _
                                                                                    b(ix + 2) * 256UI + b(ix + 3)
                                    ix += 4
                            End Select
                            p.ConvertValueFromHardwareFormat()
                            p.WriteValueToMMF()
                        End If
                    Next
                End If
            End If
        Next
        ' --------------------------------------------------------  do not charge the cpu if not exchanged
        If Not ExchangedFlag Then
            Hid.USB_Write_Read()
        End If
    End Sub


    Private Sub FastDataExchange()
        ' -------------------------------------------------------- fast data exchange
        Hid.USB_TxData(0) = Cmd.FAST_DATA_EXCHANGE

        ' -------------------------------------------------------- MMF to USB
        For Each s As Slave In Slaves
            For Each p As Pin In s.Pins
                If p.Direction = Pin.Directions.HostToMaster Then
                    p.ReadValueFromMMF()
                    p.ConvertValueToHardwareFormat()
                    p.WriteValueToUsbBuffer()
                End If
            Next
        Next

        ' -------------------------------------------------------- USB <> Master <> Slaves
        'Hid.disable_for_test = True
        Hid.USB_Write_Read()
        'Hid.disable_for_test = False
        Hid.USB_TxData(0) = Cmd.NOCMD

        ' -------------------------------------------------------- manage errors
        If Hid.USB_RxData(0) = 0 Then
            ' ---------------------------------------------------- USB to MMF
            For Each s As Slave In Slaves
                For Each p As Pin In s.Pins
                    If p.Direction = Pin.Directions.MasterToHost Then
                        p.ReadValueFromUsbBuffer()
                        p.ConvertValueFromHardwareFormat()
                        p.WriteValueToMMF()
                    End If
                Next
            Next
            Execution_OK()
        Else
            Execution_Error("FastDataExchange")
        End If

    End Sub

    Friend Function GetValuesFromSlave(ByVal SlaveId As Int32, ByRef ByteArray() As Byte) As Boolean
        ' ----------------------------------------------------
        ' SyncLock not needed because called from timer
        ' ---------------------------------------------------- send CMD / SlaveId / Nbytes
        Hid.USB_TxData(0) = Cmd.GET_VALUES
        Hid.USB_TxData(1) = CByte(SlaveId)
        Hid.USB_TxData(2) = CByte(ByteArray.Length)
        ' ---------------------------------------------------- send and receive
        Hid.USB_Write_Read()
        Hid.USB_TxData(0) = Cmd.NOCMD
        ' ---------------------------------------------------- get bytes
        For i As Int32 = 0 To ByteArray.Length - 1
            ByteArray(i) = Hid.USB_RxData(i + 1)
        Next
        ' ---------------------------------------------------- manage errors
        If Hid.USB_RxData(0) <> 0 Then
            Execution_Error("GetValuesFromSlave")
            Return False
        Else
            Execution_OK()
        End If
        Return True
    End Function

    Friend Sub SendValuesToSlave(ByVal SlaveId As Int32, ByRef bytes() As Byte)
        ' ----------------------------------------------------
        ' SyncLock not needed because called from timer
        ' ---------------------------------------------------- send CMD / SlaveId / Nbytes
        Hid.USB_TxData(0) = Cmd.SEND_VALUES
        Hid.USB_TxData(1) = CByte(SlaveId)
        Hid.USB_TxData(2) = CByte(bytes.Length)
        ' ---------------------------------------------------- send bytes
        For i As Int32 = 0 To bytes.Length - 1
            Hid.USB_TxData(3 + i) = bytes(i)
        Next
        ' ---------------------------------------------------- send and receive
        Hid.USB_Write_Read()
        Hid.USB_TxData(0) = Cmd.NOCMD
        ' ---------------------------------------------------- manage errors
        If Hid.USB_RxData(0) <> 0 Then
            Execution_Error("SendValuesToSlave")
        Else
            Execution_OK()
        End If
    End Sub



    Friend Sub GetBytesFromSlave(ByVal SlaveId As Int32, ByRef ByteArray() As Byte)
        DataExchangeThread_Stop()
        SyncLock ActionLock
            ' ---------------------------------------------------- send CMD / SlaveId / Nbytes
            Hid.USB_TxData(0) = Cmd.GET_BYTES
            Hid.USB_TxData(1) = CByte(SlaveId)
            Hid.USB_TxData(2) = CByte(ByteArray.Length)
            ' ---------------------------------------------------- send and receive
            Hid.USB_Write_Read()
            Hid.USB_TxData(0) = Cmd.NOCMD
            ' ---------------------------------------------------- get bytes
            For i As Int32 = 0 To ByteArray.Length - 1
                ByteArray(i) = Hid.USB_RxData(i + 1)
            Next
            ' ---------------------------------------------------- manage errors
            If Hid.USB_RxData(0) <> 0 Then
                Execution_Error("GetBytesFromSlave")
            Else
                Execution_OK()
            End If
        End SyncLock
        DataExchangeThread_Start()
    End Sub

    Friend Sub SendBytesToSlave(ByVal SlaveId As Int32, ByRef bytes() As Byte)
        DataExchangeThread_Stop()
        SyncLock ActionLock
            ' ---------------------------------------------------- send CMD / SlaveId / Nbytes
            Hid.USB_TxData(0) = Cmd.SEND_BYTES
            Hid.USB_TxData(1) = CByte(SlaveId)
            Hid.USB_TxData(2) = CByte(bytes.Length)
            ' ---------------------------------------------------- send bytes
            For i As Int32 = 0 To bytes.Length - 1
                Hid.USB_TxData(3 + i) = bytes(i)
            Next
            ' ---------------------------------------------------- send and receive
            Hid.USB_Write_Read()
            Hid.USB_TxData(0) = Cmd.NOCMD
            ' ---------------------------------------------------- manage errors
            If Hid.USB_RxData(0) <> 0 Then
                Execution_Error("SendBytesToSlave")
            Else
                Execution_OK()
            End If
        End SyncLock
        DataExchangeThread_Start()
    End Sub


    Private Sub Execution_Error(ByVal ErrorString As String)
        ErrorRate += 0.01F
        Debug.Print("ERROR - " & ErrorString)
    End Sub

    Private Sub Execution_OK()
        ErrorRate *= 0.95F
    End Sub


End Class

